"""
taper (:mod:`skrf.taper`)
=========================

Taper Objects

Tapered transformers, or tapers in short, are used to match
one impedance to another (from Z1 to Z2) [#]_.

By tapering a transmission line, a very broadband impedance match (low VSWR)
can be realized over a wide bandwidth, the longer the taper, the wider the frequency band.

References
----------
.. [#] https://www.microwaves101.com/encyclopedias/tapered-transformers

.. autosummary::
    :toctree: generated/

    Taper1D
    Linear
    Exponential
    SmoothStep
    Klopfenstein

"""
from __future__ import annotations

from numbers import Number
from typing import Callable

from numpy import exp, linspace, log

from .network import cascade_list


class Taper1D:
    """
    Generic 1D Taper Object
    """
    def __init__(self, med, start: Number, stop: Number, n_sections: int, f: Callable,
                 length: Number, length_unit: str = 'm', param: str = 'z0', f_is_normed: bool = True,
                 med_kw: dict = None, f_kw: dict = None):
        """
        Generic 1D Taper Constructor.

        Parameters
        ----------
        med : :class:`~skrf.media.media.Media`
            A :class:`~skrf.media.media.Media` object or a `@classmethod` `__init__`,
            used to generate the transmission line.
            See `med_kw` for arguments.
            Examples:

            * `skrf.media.RectangularWaveguide`  # a class
            * `skrf.media.RectangularWaveguide.from_z0`  # an init

        start : number
            starting value for `param`
        stop : number
            stop value for `param`
        n_sections : int
            number of sections in taper
        f : Callable
            function defining the taper transition. must take either
            no arguments or take `(x,length, start, stop)`.
            see `f_is_normed` arguments
        length : number
            physical length of the taper (in `length_unit`)
        length_unit : str
            unit of length variable. see `skrf.to_meters`
        param : str
            name of the parameter of `med` that varies along the taper
        f_is_normed: bool
            is `f` scalable and normalized. ie can f just be scaled
            to fit different start/stop conditions? if so then f is
            called with no arguments, and must  have domain and raings
            of [0,1], and [0,1]
        f_kw : dict
            passed to `f()` when  called
        med_kw : dict
            passed to `med.__init__` when an instance is created

        Note
        ----
        The default behaviour should is to taper based on impedance.
        To do this we inspect the `med` class for a `from_z0`
        init method, and if it exists, we assign it to `med` attribute,
        in `__init__`.
        Admittedly having `med` be a class or a method is abuse,
        it makes for a intuitive operation.

        Examples
        --------
        Create a linear taper from 100 to 1000mil

        >>> from skrf import Frequency, RectangularWaveguide, Taper1D, mil, inch
        >>> taper = Taper1D(med=RectangularWaveguide,
                            param='a',
                            start=100*mil,
                            stop=1000*mil,
                            length=1*inch,
                            n_sections=20,
                            f=lambda x: x,
                            f_is_normed=True,
                            med_kw={'frequency':Frequency(75,110,101,'ghz')})
        """
        if f_kw is None:
            f_kw = {}
        if med_kw is None:
            med_kw = {}
        self.med = med
        self.param = param
        self.start = start
        self.stop = stop
        self.f = f
        self.f_is_normed = f_is_normed
        self.length = length
        self.length_unit = length_unit
        self.n_sections = n_sections
        self.med_kw = med_kw
        self.f_kw = f_kw

        # the default behaviour should be to taper based on impedance.
        # to do this we inspect the media class for a `from_z0`
        # init method, and if it exists, we assign it to `med` attribute
        # admittedly having `med` be a class or a method is abuse,
        # it makes for a intuitive operation
        if param == 'z0':
            if hasattr(self.med, 'from_z0'):
                self.med = self.med.from_z0

    def __str__(self) -> str:
        return 'Taper: {classname}: {param} from {start}-{stop}'


    @property
    def section_length(self) -> Number:
        """
        Section length.

        Returns
        -------
        l : number
        """
        return  1.0*self.length/self.n_sections

    @property
    def value_vector(self):
        if self.f_is_normed:
            x = linspace(0,1,self.n_sections)
            y = self.f(x, **self.f_kw)*(self.stop-self.start) + self.start
        else:
            x = linspace(0,self.length,self.n_sections)
            y = self.f(x,self.length, self.start, self.stop, **self.f_kw)
        return y

    def media_at(self, val: Number):
        """
        Create a media instance for the taper with parameter value `val`.

        Parameters
        ----------
        val : number
            parameter value

        Returns
        -------
        media : :class:`~skrf.media.media.Media`
            media instance for the taper for the given parameter value
        """
        med_kw = self.med_kw.copy()
        med_kw.update({self.param: val})
        return self.med(**med_kw)

    def section_at(self, val: Number):
        """
        Create a single section of the taper with parameter value `val`.

        Parameters
        ----------
        val : number
            parameter value

        Returns
        -------
        media : :class:`~skrf.network.Network`
            Network instance for the section of the taper
            for the given parameter value
        """
        return self.media_at(val).line(self.section_length,
                                       unit=self.length_unit)

    @property
    def medias(self) -> list:
        """
        List of medias.

        Returns
        -------
        medias : list of :class:`~skrf.media.media.Media`
        """
        return [self.media_at(k) for k in self.value_vector]

    @property
    def sections(self) -> list:
        """
        List of sections.

        Returns
        -------
        sections : list of :class:`~skrf.network.Network`
        """
        return [self.section_at(k) for k in self.value_vector]

    @property
    def network(self):
        """
        Resulting Network

        Returns
        -------
        ntwk : :class:`~skrf.network.Network`
        """
        return cascade_list(self.sections)



class Linear(Taper1D):
    """
    A linear Taper.

    Defined by :math:`f(x)=x`

    """
    def __init__(self, **kw):
        opts = dict(f=lambda x:x, f_is_normed=True)
        kw.update(opts)
        super().__init__(**kw)


class Exponential(Taper1D):
    r"""
    An Exponential Taper.

    Defined by :math:`f(x) = f_0 \exp\left[ \frac{x}{x_1}  \ln\left( \frac{f_1}{f_0} \right) \right]`

    where:

    *    :math:`f_0`: start param value
    *    :math:`f_1`: stop param value
    *    :math:`x`: independent variable (position along taper)
    *    :math:`x_1`: length of taper

    """
    def __init__(self, **kw):
        """
        Exponential Taper Constructor
        """

        def f(x, length, start, stop):
            return start*exp(x/length*(log(stop/start)))

        opts = dict(f=f, f_is_normed=False)
        kw.update(opts)
        super().__init__(**kw)


class SmoothStep(Taper1D):
    """
    A smoothstep Taper.

    There is no analytical basis for this in the EE world that I know
    of. it is just a reasonable smooth curve, that is easy to implement.

    :math:`f(x) = (3 x^2 - 2x^3)`

    References
    ----------
    https://en.wikipedia.org/wiki/Smoothstep

    """
    def __init__(self, **kw):
        """
        Smoothstep Taper Constructor.
        """

        def f(x):
            return 3 * x ** 2 - 2 * x ** 3
        opts = dict(f=f, f_is_normed=True)
        kw.update(opts)
        super().__init__(**kw)


class Klopfenstein(Taper1D):
    """
    Klopfenstein Taper.

    This impedance taper was first described by R. W. Klopfenstein [#microwaves101]_ in a paper titled
    "A Transmission Line Taper of Improved Design", published in 1956 [#Klopfenstein]_ .
    A correction to Klopfenstein's math was published by in May 1973 by Darko Kajfez and Jame Prewit [#Kajfez]_ .

    References
    ----------
    .. [#microwaves101] https://www.microwaves101.com/encyclopedias/klopfenstein-taper
    .. [#Klopfenstein] R. W. Klopfenstein Proc. IRE. vol. 44 pp. 31-35 Jan. 1956.
        doi: 10.1109/JRPROC.1956.274847
        https://ieeexplore.ieee.org/document/4051841
    .. [#Kajfez] D. Kajfez and J. O. Prewitt, "Correction to "A Transmission Line Taper of Improved Design" (Letters),"
        in IEEE Transactions on Microwave Theory and Techniques, vol. 21, no. 5, pp. 364-364, May 1973,
        doi: 10.1109/TMTT.1973.1128003.
        https://ieeexplore.ieee.org/document/1128003
    """
    def __init__(self, **kw):
        raise NotImplementedError('Not Yet Implemented')



_scpi_error_codes_and_messages = {
    # Command Errors
    -100: ("COMMAND_ERR", "A generic command error occured"),
    -101: ("INVALID_CHARACTER", "Command contains an invalid character"),
    -102: ("SYNTAX_ERR", "An unrecognized command was encountered"),
    -103: ("INVALID_SEPERATOR", "A seperator was expected but not passed"),
    -104: ("DATA_TYPE_ERR", "An invalid data type was passed"),
    -105: ("GET_NOT_ALLOWED", "This command does not accept queries"),
    -108: ("PARAM_NOT_ALLOWED", "Too many parameters were passed"),
    -109: ("MISSING_PARAM", "One or more parameters are missing"),
    -110: ("COMMAND_HEADER_ERR", "A generic error was deteced in the header"),
    -111: ("HEADER_SEPERATOR_ERR", "An invalid seperator was encountered in the header"),
    -112: ("PRGM_MNEMONIC_TOO_LONG", "The header contains more too many characters (> 12)"),
    -113: ("UNDEFINED_HEADER", "The command is unrecognized"),
    -114: ("HEADER_SUFFIX_OUT_OF_RANGE", "The parameter passed is out of range for this header"),
    -115: ("UNEXPECTED_NUMBER_OF_PARAMS", "The number of parameters received does not correspond to the number expected"),  # noqa: E501
    -120: ('NUMERIC_DATA_ERROR', 'A generic numeric data error occurred'),
    -121: ('INVALID_CHAR_IN_NUMBER', 'An invalid character was encounterd for the numeric data type'),
    -123: ('EXPONENT_TOO_LARGE', 'The magnitude of the exponent was larger than 32,000'),
    -124: ('DECIMAL_TOO_LONG', 'Decimal data contained too many digits'),
    -128: ('NUMERIC_NOT_ALLOWED', 'Numeric data is not accepted in this context'),
    -130: ('SUFFIX_ERROR', 'A generic error occured parsing the command suffix'),
    -131: ('BAD_SUFFIX', 'There is a syntax error in the suffix or it is inappropriate for this device'),
    -134: ('SUFFIX_TOO_LONG', 'The suffix contains too many characters (> 12)'),
    -138: ('SUFFIX_NOT_ALLOWED', 'A suffix is not allowed in this context'),
    -140: ('CHAR_DATA_ERROR', 'A generic character error occurred'),
    -141: ('INVALID_CHAR_DATA', 'An invalid character was encountered'),
    -144: ('CHAR_DATA_TOO_LONG', 'The character data contains too many characters (> 12)'),
    -148: ('CHAR_NOT_ALLOWED', 'A character is not allowed in this context'),
    -150: ('STRING_DATA', 'A generic string error occurred'),
    -151: ('STRING_INVALID', 'The string is invalid'),
    -158: ('STRING_NOT_ALLOWED', 'A string is not allower in this context'),
    -160: ('BLOCK_DATA', 'A generic block data occurred'),
    -161: ('BAD_BLOCK', 'An invalid data block was encountered'),
    -168: ('BLOCK_NOT_ALLOWED', 'Block data is not allowed in this context'),
    -170: ('EXPR_ERROR', 'A generic expression error occurred'),
    -171: ('INVALID_EXPRESSION', 'The expression data element was invalid'),
    -178: ('EXPR_NOT_ALLOWED', 'An expression is not allowed in this context'),
    -180: ('MACRO_ERROR', 'A generic macro error occurred'),
    -181: ('VALID_ONLY_INSIDE_MACRO', 'A macro placeholder was encountered outside of a macro definition'),
    -183: ('INVALID_WITHIN_MACRO', 'There is a syntax error in the macro'),
    -184: ('MACRO_PARM', 'A command in the macro had the wrong number or type of parameters'),
    # Execution Errors
    -200: ('EXEC_ERROR', 'A generic execution error occurred.'),
    -201: ('INVALID_WHILE_IN_LOCAL', 'The command is not executable when the device is in local'),
    -202: ('SETTINGS_LOST', 'A setting was lost due to RTL'),
    -203: ('COMMAND_PROTECTED', 'The command is password protected or disabled'),
    -210: ('TRIGGER_ERROR', 'A generic trigger error occurred'),
    -211: ('TRIGGER_IGNORED', 'The device ignored the command'),
    -212: ('ARM_IGNORED', 'The device ignored the arming signal'),
    -213: ('INIT_IGNORED', 'The device ignored the init command because another measurement was in progress'),
    -214: ('TRIGGER_DEADLOCK', 'The trigger is in a deadlocked state'),
    -215: ('ARM_DEADLOCK', 'The arming source is in a deadlocked state'),
    -220: ('PARAM_ERROR', 'A generic parameter error occurred'),
    -221: ('SETTINGS_CONFLICT', 'The command could not be executed because of a settings conflict'),
    -222: ('DATA_OUT_OF_RANGE', 'The interpreted value is outside the legal range'),
    -223: ('TOO_MUCH_DATA', 'The device could not handle all of the data'),
    -224: ('ILLEGAL_PARAM_VALUE', 'An invalid parameter was passed'),
    -225: ('OUT_OF_MEMORY', 'The device has insufficient memory to perform the command'),
    -226: ('LIST_LENGTH', 'Lists not of equal lengths'),
    -230: ('DATA_CORRUPT_OR_STALE', 'Data is invalid'),
    -231: ('DATA_QUESTIONABLE', 'The measurement accuracy is suspect'),
    -232: ('INVALID_FORMAT', 'The data is of an invalid format'),
    -233: ('INVALID_VERSION', 'The data is of an invalid version'),
    -240: ('HARDWARE_ERROR', 'A generic hardware error occurred'),
    -241: ('HARDWARE_MISSING', 'The command could not be executed because of missing device hardware'),
    -250: ('MASS_STORAGE_ERROR', 'A generic mass storage error occurred'),
    -251: ('MISSING_MASS_STORAGE', 'There is no mass storage to execute this command'),
    -252: ('MISSING_MEDIA', 'Media required to run this command is not present'),
    -253: ('CORRUPT_MEDIA', 'Media required to run this command is corrupt'),
    -254: ('MEDIA_FULL', 'Media required to run this command is full'),
    -255: ('DIRECTORY_FULL', 'The specified directory is full'),
    -256: ('FILE_NOT_FOUND', 'The file was not found'),
    -257: ('FILE_NAME_ERROR', 'Encountered an error in the filename'),
    -258: ('MEDIA_PROTECTED', 'The command could not be executed because the media is protected'),
    -260: ('EXPRESSION_ERROR', 'An error occurred with an expression program data element'),
    -261: ('MATH_ERROR', 'A mathematical error occrred'),
    -270: ('MACRO_EXECUTION', 'An error occurred when executing a macro'),
    -271: ('MACRO_SYNTAX', 'There was a syntax error in the macro'),
    -272: ('MACRO_EXEC', 'There was an error in the macro definition'),
    -273: ('BAD_MACRO_NAME', 'The macro has an invalid name'),
    -274: ('MACRO_PARAM_ERROR', 'The macro definition improperly used a macro parameter placeholder'),
    -275: ('MACRO_TOO_LONG', 'The macro was too long to for the device to handle'),
    -276: ('MACRO_RECURSION', 'The macro was infinitely recursive'),
    -277: ('CANT_REDEFINE_MACRO', 'The macro label is already defined'),
    -278: ('MACRO_NOT_FOUND', 'The specified macro is not defined'),
    -280: ('PROGRAM_ERROR', 'A generic program error occurred'),
    -281: ('CANT_CREATE_PROGRAM', 'An attempt to create a program was unsuccesful'),
    -282: ('ILLEGAL_PROGRAM_NAME', 'The program name is invalid'),
    -283: ('ILLEGAL_VAR_NAME', 'The variable does not exist'),
    -284: ('PROGRAM_RUNNING', 'The operation is illegal while the program is running'),
    -285: ('PROGRAM_SYNTAX', 'A syntax error was encountered in the program'),
    -286: ('PROGRAM_RUNTIME', 'A program runtime error occurred'),
    -290: ('MEMORY_USE', 'A user request has resulted in a memory error'),
    -291: ('OUT_OF_MEMORY', 'Out of memory'),
    -292: ('NAME_NOT_FOUND', 'The referenced name does not exist'),
    -293: ('NAME_ALREADY_EXISTS', 'The referenced name already exists'),
    -294: ('INCOMPATIBLE_TYPE', 'The type or structure of the memory item is inadequate'),
    # Device specific errors
    -300: ('DEVICE_SPECIFIC_ERROR', 'A generic device error occurred'),
    -310: ('SYSTEM_ERROR', 'A system error occurred'),
    -311: ('MEMORY_ERROR', 'A physical fault in the device`s memory occurred'),
    -312: ('PUD_MEMORY_LOST', 'Protected user data saved by a *PUD command was lost'),
    -313: ('CAL_MEMORY_LOST', 'Calibration data saved by a *CAL? command was lost'),
    -314: ('SAV_RCL_MEMORY_LOST', 'Data saved by the *SAV? command has been lost'),
    -315: ('CONFIG_MEMORY_LOST', 'Configuration data has been lost'),
    -320: ('STORAGE_FAULT', 'The firmware detected a fault when using data storage'),
    -321: ('OUT_OF_MEMORY', 'An internal operation needed more memory than was available'),
    -330: ('SELF_TEST_FAILED', 'The instrument self test failed'),
    -340: ('CAL_FAILED', 'The device experienced a calibration error'),
    -350: ('QUEUE_OVERFLOW', 'An error occurred, but there were too many errors in the error queue to record it'),
    -360: ('COMMUNICATION_ERROR', 'A generic communication error occurred'),
    -361: ('PARITY_ERROR', 'The received data had an invalid parity bit'),
    -362: ('FRAMING_ERROR', 'A stop bit was not detected'),
    -363: ('INPUT_BUFFER_OVERRUN', 'The serial port input buffers overflowed'),
    -365: ('TIMEOUT_ERROR', 'A timeout error occurred'),
    # Query errors
    -400: ('QUERY_ERROR', 'A generic query error occurred'),
    -410: ('QUERY_INTERRUPTED', 'An interrupted query occurred'),
    -420: ('QUERY_UNTERMINATED', 'The query was unterminated'),
    -430: ('QUERY_DEADLOCKED', 'The query encountered a deadlock error'),
    -440: ('RESPONSE_NOT_ALLOWED', 'A query was received after an indefinite query was executed'),
    # Other errors
    -500: ('POWER_ON', 'The device detected an off to on transition in its power supply'),
    -600: ('USER_REQUEST', 'The device detected the activation of a user request local control'),
    -700: ('REQUEST_CONTROL', 'The device requested to be the active IEEE 488.1 controller-in-charge'),
    -800: ('OPERATION_COMPLETE', 'The instrument has completed all pending operations'),
}


class SCPIError(Exception):
    def __init__(self, errno: int) -> None:
        abbreviation, description = _scpi_error_codes_and_messages.get(
            errno, ("UNKNOWN_ERR", "An unknown error occurred.")
        )

        super().__init__(
            f"{abbreviation} ({errno}): {description}"
        )
        self.error_code = errno
        self.abbreviation = abbreviation
        self.description = description



"""
DefinedAEpTandZ0 (:mod:`skrf.media.definedAEpTandZ0`)
======================================================

Transmission line medium defined by A, Ep, Tand and Z0.

This medium is defined by attenuation A, relative permittivity Ep_r,
loss angle tand and characteristic impedance Z0.

Djirdjevic/Svennson dispersion model is provided for dielectric, default
behaviour is frequency invariant.

.. autosummary::
   :toctree: generated/

   DefinedAEpTandZ0
"""
from __future__ import annotations

import warnings
from typing import Sequence

from numpy import imag, log, ndarray, ones, pi, real, sqrt
from scipy.constants import c

from ..constants import NumberLike
from ..frequency import Frequency
from .media import Media


class DefinedAEpTandZ0(Media):
    r"""
    Transmission line medium defined by A, Ep, Tand and Z0.

    This medium is defined by attenuation `A`, relative permittivity `Ep_r`,
    loss angle `tand` and nominal impedance `Z0`.

    Djirdjevic [#Djordjevic]_ / Svennson [#Svensson]_ dispersion model
    is provided for dielectric. Default behaviour is frequency invariant.

    Parameters
    ----------
    frequency : :class:`~skrf.frequency.Frequency` object
        Frequency band of this transmission line medium
    z0_port : number, array-like, or None
        `z0_port` is the port impedance for networks generated by the media.
        If `z0_port` is not None, the networks generated by the media are
        renormalized (or in other words embedded) from the characteristic
        impedance z0 of the media to `z0_port`.
        Else if `z0_port` is None, the networks port impedances will be the raw
        characteristic impedance z0 of the media.
        (Default is None)
    A : number, array-like, default 0.0
        Attenuation due to conductor loss in dB/m/sqrt(Hz)
        The attenuation :math:`A(f)` at frequency :math:`f` is:

        .. math::

            A(f) = A\sqrt{\frac{f}{f_A}}

    f_A : number, default 1.0
        Frequency scaling in Hz for the attenuation. See A.
    ep_r : number, array-like, default 1.0
        Real part of the relative permittivity of the dielectric:
        :math:`\epsilon_r'=\Re[\epsilon]`.

        If `model='frequencyinvariant'`, the complex relative permittivity is:

        .. math::

            \epsilon_r(f) = \epsilon'_r + j \cdot \epsilon'_r \cdot \tan\delta

        if `model='djordjevicsvensson'`, the complex relative permittivity is:

        .. math::

            \epsilon_r(f) = \epsilon_\inf + m \cdot \ln{\frac{f_{high} + j f_\epsilon}{f_{low} + j f_\epsilon}}

        where :math:`\epsilon_\inf` is the permittivity value when frequency approaches infinity.
        In this case, the value of :math:`\epsilon_r` and :math:`\tan\delta` (tanD)
        are given for frequency :math:`f_\epsilon`:

        .. math::

            \epsilon_r(f_\epsilon) = \epsilon'_r+ j \cdot \epsilon'_r \cdot \tan\delta

    tanD : number, array-like, default 0.0
        Dielectric relative permittivity loss tangent :math:`\tan\delta`. See `ep_r`.
    z0 : number, array-like, default 50.0
        Quasi-static nominal impedance of the medium.
        Because of the dispersion introduced by the conductor and dielectric
        losses, the characteristic impedance can be frequency-dependent with an
        imaginary part. The characteristic impedance is computed with an RLGC
        model.
        If the impedance parameter is array-like, the characteristic impedance
        is assigned to this value without modification.

        .. math::

            R = 2 Z_n \alpha_{conductor}

        .. math::

            L = \frac{Z_n \sqrt{\epsilon_r}}{C_0}

        .. math::

            G = \frac{2}{Z_n} \alpha_{dielectric}

        .. math::

            C = \frac{\sqrt{\epsilon_r}}{C_0 Z_n}

        .. math::

            Z_0 = \sqrt{\frac{R + j\omega L}{G + j\omega C}}

        where :math:`Z_n` is the nominal impedance and :math:`Z_0` is the
        characteristic impedance.
    Z0 : number, array-like, or None
        deprecated parameter, only emmit a deprecation warning.
    f_low : number, default 1e3, optional
        Low frequency in Hz for  for Djordjevic/Svennson dispersion model.
        See `ep_r`.
    f_high : number, default 1e12, optional
        High frequency in Hz for for Djordjevic/Svennson dispersion model.
        See `ep_r`.
    f_ep : number, default 1e9, , optional
        Specification frequency in Hz for  for Djirdjevic/Svennson dispersion model.
        `ep_r` and `tanD` parameters are specified for this frequency. See `ep_r`.
    model : string, 'frequencyinvariant' or 'djordjevicsvensson', optional
        Use Djirdjevic/Svennson wideband Debye dispersion model or not.
        Default is frequency invariant behaviour.
    \*args, \*\*kwargs : arguments and keyword arguments


    Examples
    --------
    >>> from skrf.media.definedAEpTandZ0 import DefinedAEpTandZ0
    >>> from skrf.frequency import Frequency
    >>> f = Frequency(75,110,101,'ghz')
    >>> DefinedAEpTandZ0(frequency=f, A=1, f_A=1e9, ep_r=3, z0=50)
    >>> DefinedAEpTandZ0(frequency=f, A=1, f_A=1e9, ep_r=3, tand=0.02, z0=50)
    >>> DefinedAEpTandZ0(frequency=f, A=1, f_A=1e9, ep_r=3, tand=0.02, z0=50,
                        f_low=1e3, f_high=1e12, f_Ep=1e9,
                        model='djordjevicsvensson')

    References
    ----------
    .. [#Svensson] C. Svensson and G. E. Dermer, "Time Domain Modeling of Lossy Interconnects,"
        IEEE Trans. Advanced Packaging, Vol. 24, No. 2, May 2001.
        https://doi.org/10.1109/6040.928754
    .. [#Djordjevic] A. R. Djordjevic, R. M. Biljic, V. D. Likar-Smiljanic, and T. K. Sarkar,
        "Wideband Frequency-Domain Characterization of FR-4 and Time-Domain Causality,"
        IEEE Trans. Electromagnetic Compatibility, Vol. 43, No. 4, November 2001.
        https://doi.org/10.1109/15.974647

    """

    def __init__(self, frequency: Frequency| None = None,
                 A: float = 0.0, f_A: float = 1.0,
                 ep_r: NumberLike = 1.0, tanD: NumberLike = 0.0,
                 z0_port: NumberLike | None = None,
                 z0: float = 50.0,
                 Z0: NumberLike | None = None,
                 f_low: float = 1.0e3, f_high: float = 1.0e12, f_ep: float = 1.0e9,
                 model: str = 'frequencyinvariant', *args, **kwargs):

        Media.__init__(self, frequency=frequency, z0_port=z0_port)
        self.A, self.f_A = A, f_A
        self.ep_r, self.tanD = ep_r, tanD

        self.f_low, self.f_high, self.f_ep = f_low, f_high, f_ep
        self.model = model

        if isinstance(z0, (Sequence, ndarray)):
            # keep raw impedance for characteristic impedance
            self.z0_characteristic = z0
        else:
            # z0 is the nominal impedance. Compute characteristic impedance with
            # an RLGC model based on alpha conductor and alpha dielectric.
            self.Zn = z0
            self.R = 2 * self.Zn * self.alpha_conductor
            self.L = self.Zn * sqrt(self.ep_r) / c
            self.G = 2 / self.Zn * self.alpha_dielectric
            self.C = sqrt(ep_r) / c / self.Zn
            self.z0_characteristic = sqrt(
                (self.R + 1j * self.frequency.w * self.L) /
                (self.G + 1j * self.frequency.w * self.C))

        if Z0 is not None:
            # warns of deprecation
            warnings.warn(
                'Use of `Z0` in DefinedAEpTandZ0 initialization is deprecated.\n'
                '`Z0` has no effect. Use `z0` instead\n'
                '`Z0` will be removed in version 1.0',
              DeprecationWarning, stacklevel = 2)

    def __str__(self):
        f = self.frequency
        output = 'DefinedAEpTandZ0 medium.  %i-%i %s.  %i points'%\
            (f.f_scaled[0], f.f_scaled[-1], f.unit, f.npoints)
        return output

    def __repr__(self):
        return self.__str__()

    @property
    def ep_r_f(self) -> NumberLike:
        """
        Frequency dependent complex relative permittivity of dielectric.
        """
        ep_r, tand  = self.ep_r, self.tanD
        f_low, f_high, f_ep = self.f_low, self.f_high, self.f_ep
        f = self.frequency.f
        if self.model == 'djordjevicsvensson':
           # compute the slope for a log frequency scale, tanD dependent.
           k = log((f_high + 1j * f_ep) / (f_low + 1j * f_ep))
           fd = log((f_high + 1j * f) / (f_low + 1j * f))
           ep_d = -tand * ep_r  / imag(k)
           # value for frequency above f_high
           ep_inf = ep_r * (1. + tand * real(k) / imag(k))
           # compute complex permitivity
           return ep_inf + ep_d * fd
        elif self.model == 'frequencyinvariant':
            return ones(self.frequency.f.shape) * (ep_r - 1j*ep_r*tand)
        else:
            raise ValueError('Unknown dielectric dispersion model')

    @property
    def tand_f(self) -> NumberLike:
        """
        Frequency dependent dielectric loss factor.
        """
        ep_r = self.ep_r_f
        return -imag(ep_r) / real(ep_r)

    @property
    def alpha_conductor(self) -> NumberLike:
        """
        Losses due to conductor resistivity.
        """
        A, f_A, f = self.A, self.f_A, self.frequency.f
        return A * log(10)/20 * sqrt(f/f_A)

    @property
    def alpha_dielectric(self) -> NumberLike:
        """
        Losses due to dielectric
        """
        ep_r, tand = real(self.ep_r_f), self.tand_f
        f = self.frequency.f
        return pi*sqrt(ep_r)*f/c * tand

    @property
    def beta_phase(self) -> NumberLike:
        """
        Phase parameter
        """
        ep_r, f = real(self.ep_r_f), self.frequency.f
        return 2*pi*f*sqrt(ep_r)/c

    @property
    def gamma(self) -> NumberLike:
        """
        Propagation constant.

        See Also
        --------
        alpha_conductor : calculates conductor losses
        alpha_dielectric: calculates dielectric losses
        beta_phase      : calculates phase parameter
        """
        beta  = self.beta_phase
        alpha = self.alpha_conductor + self.alpha_dielectric
        return alpha + 1j*beta

    @property
    def z0_characteristic(self) -> NumberLike:
        """
        Characteristic Impedance, :math:`z_0`

        Returns
        -------
        z0_characteristic : np.ndarray
            Characteristic Impedance in units of ohms
        """
        return self._z0

    @z0_characteristic.setter
    def z0_characteristic(self, val):
        self._z0 = val



import sys

import numpy as np
import pytest

import skrf

try:
    from skrf.vi.vna import ValuesFormat, keysight
    from skrf.vi.vna.keysight.fieldfox import WindowFormat
except ImportError:
    pytest.skip(allow_module_level=True)

if "matplotlib" not in sys.modules:
    pytest.skip(allow_module_level=True)

@pytest.fixture
def mocked_ff(mocker):
    mocker.patch('skrf.vi.vna.keysight.FieldFox.__init__', return_value=None)
    mocker.patch('skrf.vi.vna.keysight.FieldFox.write')
    mocker.patch('skrf.vi.vna.keysight.FieldFox.write_values')
    mocker.patch('skrf.vi.vna.keysight.FieldFox.query')
    mocker.patch('skrf.vi.vna.keysight.FieldFox.query_values')

    yield keysight.FieldFox('TEST')

@pytest.mark.parametrize(
    'param,expected_query,expected_write,query_response,expected_val,write_val',
    [
        ('freq_start', 'SENS:FREQ:STAR?', 'SENS:FREQ:STAR 100', '100', 100, 100),
        ('freq_stop', 'SENS:FREQ:STOP?', 'SENS:FREQ:STOP 100', '100', 100, 100),
        ('freq_center', 'SENS:FREQ:CENT?', 'SENS:FREQ:CENT 100', '100', 100, 100),
        ('freq_span', 'SENS:FREQ:SPAN?', 'SENS:FREQ:SPAN 100', '100', 100, 100),
        ('npoints', 'SENS:SWE:POIN?', 'SENS:SWE:POIN 100', '100', 100, 100),
        ('if_bandwidth', 'SENS:BWID?', 'SENS:BWID 100', '100', 100, 100),
        ('window_configuration', 'DISP:WIND:SPL?', 'DISP:WIND:SPL D1', 'D1',
         WindowFormat.ONE_TRACE, WindowFormat.ONE_TRACE),
        ('ntraces', 'CALC:PAR:COUN?', 'CALC:PAR:COUN 1', '1', 1, 1),
        ('active_trace', None, 'CALC:PAR1:SEL', None, None, 1),
    ]
)
def test_params(
    mocker,
    mocked_ff,
    param,
    expected_query,
    expected_write,
    query_response,
    expected_val,
    write_val
):
    if expected_query is not None:
        mocked_ff.query.return_value = query_response
        test_query = getattr(mocked_ff, param)
        mocked_ff.query.assert_called_once_with(expected_query)
        assert test_query == expected_val

    if expected_write is not None:
        setattr(mocked_ff, param, write_val)
        mocked_ff.write.assert_called_once_with(expected_write)

def test_freq_query(mocker, mocked_ff):
    mocked_ff.query.side_effect = [
        '100', '200', '11'
    ]
    test = mocked_ff.frequency
    assert test == skrf.Frequency(100, 200, 11, unit='hz')

def test_freq_write(mocker, mocked_ff):
    test_f = skrf.Frequency(100, 200, 11, unit='hz')
    mocked_ff.frequency = test_f
    calls = [
        mocker.call("SENS:FREQ:STAR 100"),
        mocker.call("SENS:FREQ:STOP 200"),
        mocker.call("SENS:SWE:POIN 11"),
    ]
    mocked_ff.write.assert_has_calls(calls)

def test_query_fmt_query(mocker, mocked_ff):
    mocked_ff.query.side_effect = ['ASC,0', 'REAL,32', 'REAL,64']
    test = mocked_ff.query_format
    assert test == ValuesFormat.ASCII
    test = mocked_ff.query_format
    assert test == ValuesFormat.BINARY_32
    test = mocked_ff.query_format
    assert test == ValuesFormat.BINARY_64

def test_query_fmt_write(mocker, mocked_ff):
    mocked_ff.query_format = ValuesFormat.ASCII
    mocked_ff.write.assert_called_with('FORM ASC,0')
    mocked_ff.query_format = ValuesFormat.BINARY_32
    mocked_ff.write.assert_called_with('FORM REAL,32')
    mocked_ff.query_format = ValuesFormat.BINARY_64
    mocked_ff.write.assert_called_with('FORM REAL,64')

def test_define_msmnt(mocker, mocked_ff):
    mocker.patch('skrf.vi.vna.keysight.FieldFox.ntraces', return_value=1, new_callable=mocker.PropertyMock)
    mocked_ff.define_measurement(1, 'S11')
    mocked_ff.write.assert_called_once_with('CALC:PAR1:DEF S11')

# def test_calibration_write(mocker, mocked_ff):
    # cal_terms = mocked_ff._cal_term_map
    # mock_array = np.array([1+1j, 1+1j])
    # cal_dict = {k: mock_array.copy() for k in cal_terms.keys()}
    # mock_cal = skrf.Calibration.from_coefs(skrf.Frequency(100, 200, 11, unit='hz'), cal_dict)
    # mocked_ff.calibration = mock_cal

    # numpy defines == in a way that makes mocker.assert_has_calls not work
    # Instead, we can just do the individual comparisons ourselves
    # expected_calls = [
        # mocker.call(f'SENS:CORR:COEF {term},', np.array([1.,1.,1.,1.]))
        # for term in cal_terms.values()
    # ]
    # actual_calls = mocked_ff.write_values.call_args_list
    # for actual, expected in zip(actual_calls, expected_calls):
        # assert actual[0][0] == expected[1][0]
        # np.testing.assert_array_almost_equal(actual[0][1], expected[1][1])

def test_get_measurement_parameter(mocker, mocked_ff):
    mocked_ff.query.return_value='S11'
    test = mocked_ff.get_measurement_parameter(1)
    assert test == 'S11'

    with pytest.raises(ValueError):
        test = mocked_ff.get_measurement_parameter(5)

def test_sweep(mocker, mocked_ff):
    mocked_ff._resource = mocker.Mock()
    mocked_ff.query.return_value='1'
    mocked_ff.sweep()
    calls = [
        mocker.call('INIT:CONT 0'),
        mocker.call('INIT'),
        mocker.call('INIT:CONT 1'),
    ]
    mocked_ff.write.assert_has_calls(calls)
    mocked_ff._resource.clear.assert_called_once()

def test_get_snp_network(mocker, mocked_ff):
    mocker.patch('skrf.vi.vna.keysight.FieldFox.sweep')
    query_ret_vals = [
        '1',
        WindowFormat.ONE_TRACE,
        '1',
        'S11',
        '4', '4', '4', '4',
        '100', '200', '11',
        '4'
    ]
    mocked_ff.query.side_effect=query_ret_vals
    mock_s_data = np.array([1. + 1.j]*11)
    mocker.patch('skrf.vi.vna.keysight.FieldFox.active_trace_sdata',
                 return_value=mock_s_data.copy(), new_callable=mocker.PropertyMock)

    test = mocked_ff.get_snp_network()

    mocked_ff.sweep.assert_called_once()
    assert isinstance(test, skrf.Network)
    assert test.s.shape == (11,2,2)
    expected = np.array([1+1j]*11)
    for s in [getattr(test, f'{param}') for param in ['s11', 's12', 's21', 's22']]:
        actual = s.s.reshape((-1,))
        np.testing.assert_array_almost_equal(actual, expected)



"""
.. module:: skrf.vi.vna.keysight.fieldfox
=================================================
Keysight FieldFox (:mod:`skrf.vi.vna.keysight.fieldfox`)
=================================================

Provides a class to interact with Keysight's FieldFox handheld analyzer.

This module additionally provides some enumerations to make certain commands
more clear.

Enums
============

.. autosummary::
    :nosignatures:
    :toctree: generated/

    WindowFormat

FieldFox Class
==============

.. autosummary::
    :nosignatures:
    :toctree: generated/

    FieldFox

"""
import itertools
from enum import Enum

import numpy as np

import skrf
from skrf.vi import vna
from skrf.vi.validators import (
    BooleanValidator,
    EnumValidator,
    FloatValidator,
    FreqValidator,
    IntValidator,
    SetValidator,
)


class WindowFormat(Enum):
    """How windows are arranged on the display."""

    #: A single trace in a window that takes up the whole display
    ONE_TRACE = "D1"
    #: Two traces in a single window
    TWO_TRACES = "D2"
    #: Three traces in a single window
    THREE_TRACES = "D3"
    #: Two windows with one trace each, stacked vertically
    TWO_VERTICAL = "D12H"
    #: Three windows. The top window is two columns wide, the bottom two are one
    #: column each
    ONE_FIRST_ROW_TWO_SECOND_ROW = "D11_23"
    #: Four windows in a grid
    TWO_BY_TWO = "D12_34"


class FieldFox(vna.VNA):
    """Keysight FieldFox.

    The FieldFox is a class of handheld analyzers with numerous modes with
    different capabilities. This class only provides methods related to the
    network analyzer mode.
    """

    freq_start = vna.VNA.command(
        get_cmd="SENS:FREQ:STAR?",
        set_cmd="SENS:FREQ:STAR <arg>",
        doc="""The start frequency [Hz]""",
        validator=FreqValidator(),
    )

    freq_stop = vna.VNA.command(
        get_cmd="SENS:FREQ:STOP?",
        set_cmd="SENS:FREQ:STOP <arg>",
        doc="""The start frequency [Hz]""",
        validator=FreqValidator(),
    )

    freq_center = vna.VNA.command(
        get_cmd="SENS:FREQ:CENT?",
        set_cmd="SENS:FREQ:CENT <arg>",
        doc="""The center frequency [Hz]""",
        validator=FreqValidator(),
    )

    freq_span = vna.VNA.command(
        get_cmd="SENS:FREQ:SPAN?",
        set_cmd="SENS:FREQ:SPAN <arg>",
        doc="""The frequency span [Hz]""",
        validator=FreqValidator(),
    )

    npoints = vna.VNA.command(
        get_cmd="SENS:SWE:POIN?",
        set_cmd="SENS:SWE:POIN <arg>",
        doc="""The number of frequency points""",
        validator=IntValidator(),
    )

    sweep_time = vna.VNA.command(
        get_cmd="SENS:SWE:TIME?",
        set_cmd="SENS:SWE:TIME <arg>",
        doc="""The sweep time [s]""",
        validator=FloatValidator(),
    )

    if_bandwidth = vna.VNA.command(
        get_cmd="SENS:BWID?",
        set_cmd="SENS:BWID <arg>",
        doc="""The center frequency [Hz]""",
        validator=SetValidator([10, 30, 100, 300, 1000, 10_000, 30_000, 100_000]),
    )

    window_configuration = vna.VNA.command(
        get_cmd="DISP:WIND:SPL?",
        set_cmd="DISP:WIND:SPL <arg>",
        doc="""How multiple trace windows appear on screen""",
        validator=EnumValidator(WindowFormat),
    )

    ntraces = vna.VNA.command(
        get_cmd="CALC:PAR:COUN?",
        set_cmd="CALC:PAR:COUN <arg>",
        doc="""The number of active traces.""",
        validator=IntValidator(min=1, max=4),
    )

    active_trace = vna.VNA.command(
        set_cmd="CALC:PAR<arg>:SEL",
        doc="""Set the active trace. There is no command to read the active
            trace.""",
        validator=IntValidator(min=1, max=4),
    )

    active_trace_sdata = vna.VNA.command(
        get_cmd="CALC:DATA:SDATA?",
        doc="""Get the current trace data as a network""",
        values=True,
        complex_values=True
    )

    is_continuous = vna.VNA.command(
        get_cmd="INIT:CONT?",
        set_cmd="INIT:CONT <arg>",
        doc="""Get the current trace data as a network""",
        validator=BooleanValidator()
    )

    _cal_term_map = {
        "forward directivity": "ed,1,1",
        "reverse directivity": "ed,2,2",
        "forward source match": "es,1,1",
        "reverse source match": "es,2,2",
        "forward reflection tracking": "er,1,1",
        "reverse reflection tracking": "er,2,2",
        "forward transmission tracking": "et,2,1",
        "reverse transmission tracking": "et,1,2",
        "forward load match": "el,2,1",
        "reverse load match": "el,1,2",
        "forward isolation": "ex,2,1",
        "reverse isolation": "ex,1,2"
    }

    def __init__(self, address: str, backend: str = "@py") -> None:
        super().__init__(address, backend)

        self._resource.read_termination = "\n"
        self._resource.write_termination = "\n"

        _ = self.query_format # calling the getter sets _values_format to make sure we're in sync with the instrument

    @property
    def nports(self) -> int:
        return 2

    @property
    def freq_step(self) -> int:
        """The frequency step [Hz]"""
        f = self.frequency
        return int(f.step)

    @freq_step.setter
    def freq_step(self, f: int) -> None:
        freq = self.frequency
        self.npoints = len(range(int(freq.start), int(freq.stop) + f, f))


    @property
    def frequency(self) -> skrf.Frequency:
        """The frequency bounds as a :class:`skrf.Frequency`"""
        return skrf.Frequency(
            start=self.freq_start, stop=self.freq_stop, npoints=self.npoints, unit='Hz'
        )

    @frequency.setter
    def frequency(self, f: skrf.Frequency):
        self.freq_start = f.start
        self.freq_stop = f.stop
        self.npoints = f.npoints

    @property
    def calibration(self) -> skrf.Calibration:
        """The currently defined calibration as a :class:`skrf.calibration.calibration.Calibration`"""
        cal_dict = {}
        for cal_key, term in self._cal_term_map.items():
            vals = self.query_values(f"SENS:CORR:COEF? {term}", container=np.array, complex_values=True)
            cal_dict[cal_key] = vals

        return skrf.Calibration.from_coefs(self.frequency, cal_dict)

    @calibration.setter
    def calibration(self, cal: skrf.Calibration) -> None:
        cal_dict = cal.coefs_12term
        for cal_key, term in self._cal_term_map.items():
            self.write_values(f"SENS:CORR:COEF {term},", cal_dict[cal_key], complex_values=True)

    @property
    def query_format(self) -> vna.ValuesFormat:
        """
        How values are written to / queried from the instrument (ascii or
        binary)

        When transferring a large number of values from the instrument (like
        trace data), it can be done either as ascii characters or as binary.

        Transferring in binary is much faster, as large numbers can be
        represented much more succinctly.
        """
        fmt = self.query("FORM?")
        if fmt == "ASC,0":
            self._values_fmt = vna.ValuesFormat.ASCII
        elif fmt == "REAL,32":
            self._values_fmt = vna.ValuesFormat.BINARY_32
        elif fmt == "REAL,64":
            self._values_fmt = vna.ValuesFormat.BINARY_64
        return self._values_fmt

    @query_format.setter
    def query_format(self, fmt: vna.ValuesFormat) -> None:
        if fmt == vna.ValuesFormat.ASCII:
            self._values_fmt = vna.ValuesFormat.ASCII
            self.write("FORM ASC,0")
        elif fmt == vna.ValuesFormat.BINARY_32:
            self._values_fmt = vna.ValuesFormat.BINARY_32
            self.write("FORM REAL,32")
        elif fmt == vna.ValuesFormat.BINARY_64:
            self._values_fmt = vna.ValuesFormat.BINARY_64
            self.write("FORM REAL,64")

    def get_measurement_parameter(self, trace: int) -> str:
        """Get the defined measurement parameter for the specified trace

        Parameters
        ----------
        trace: int
            The trace number

        Returns
        -------
        str
            Measurement parameter (like 'S11', 'S21', 'A', 'B', 'R1', etc)
        """
        if trace not in range(1, 5):
            raise ValueError("Trace must be between 1 and 4")

        return self.query(f"CALC:PAR{trace}:DEF?")

    def define_measurement(self, trace: int, parameter: str) -> None:
        """
        Define what parameter the specified trace is measuring

        Parameters
        ----------
        trace: int
            What trace to set the measurement for
        parameter: str
            What parameter to measure (e.g. 'S11', 'R1', 'A', etc)
        """
        if trace not in range(1, self.ntraces + 1):
            self.ntraces = trace

        self.write(f"CALC:PAR{trace}:DEF {parameter}")

    def sweep(self) -> None:
        """Trigger a fresh sweep."""
        self._resource.clear()
        was_continuous = self.is_continuous
        self.is_continuous = False
        self.write("INIT")
        self.is_continuous = was_continuous

    def get_snp_network(self, ports=None, restore_settings: bool = True) -> skrf.Network:
        """
        Get trace data as an :class:`skrf.Network`

        Parameters
        ----------
        ports: Sequence
            Which ports to get s parameters for. Can only be 1, 2, or (1, 2)
        restore_settings: bool
            Whether the original settings should be saved and restored, like the
            number of traces, the trace parameters, etc. Doing this takes a bit
            more time, so if running multiple repeated measurements, pass False
            (Default True)

        Returns
        -------
        :class:`skrf.Network`
            The measured data
        """
        if ports is None:
            ports = {1, 2}
        msmnts = list(itertools.product(ports, repeat=2))
        msmnt_params = [f"S{a}{b}" for a, b in msmnts]

        if restore_settings:
            original_config = {
                "ntraces": self.ntraces,
                "window_configuration": self.window_configuration,
                "trace_params": [self.get_measurement_parameter(i+1) for i in range(self.ntraces)]
            }

        self.ntraces = len(msmnts)
        for i, param in enumerate(msmnt_params):
            self.define_measurement(i+1, param)

        ntwk = skrf.Network()
        ntwk.frequency = self.frequency
        ntwk.s = np.empty(
            shape=(ntwk.frequency.npoints, len(ports), len(ports)), dtype=complex
        )

        self.sweep()
        for tr, (i, j) in enumerate(msmnts):
            self.active_trace = tr + 1

            sdata = self.active_trace_sdata
            if len(msmnts) == 1:
                ntwk.s[:, 0, 0] = sdata
            else:
                ntwk.s[:, i - 1, j - 1] = sdata

        if restore_settings:
            for i, param in enumerate(original_config.pop('trace_params')):
                self.define_measurement(i+1, param)
            for key, val in original_config.items():
                setattr(self, key, val)

        return ntwk
